# Generated by Django 4.2.16 on 2024-10-15 13:49

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("receipts", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="jobstartedreceipt",
            name="is_organic",
            field=models.BooleanField(default=False),
            preserve_default=False,
        ),
    ]
