# Generated by Django 4.2.19 on 2025-04-27 23:22

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("receipts", "0003_jobacceptedreceipt_and_more"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="jobstartedreceipt",
            name="max_timeout",
        ),
    ]
